package main

import (
	"database/sql"
	"encoding/hex"
	"fmt"
	"github.com/boramalper/magnetico/pkg/persistence"
	"github.com/cheggaaa/pb/v3"
	_ "github.com/mattn/go-sqlite3"
	"go.uber.org/zap"
	"gopkg.in/alecthomas/kingpin.v2"
	"net/url"
)

var (
	goDatabaseUrl = kingpin.Flag("go-database", "magneticod Go version database URL.").Short('g').String()
	pyDatabaseUrl = kingpin.Flag("py-database", "Python version database URL.").Short('p').Required().String()
)

func main() {
	logger, _ := zap.NewDevelopment()
	defer logger.Sync()
	zap.ReplaceGlobals(logger)

	kingpin.Parse()

	goDatabase, err := persistence.MakeDatabase(*goDatabaseUrl, logger)
	if err != nil {
		logger.Panic("Could not open Go database.", zap.String("url", *goDatabaseUrl), zap.Error(err))
	}
	defer goDatabase.Close()

	oldSqliteUrl, err := url.Parse(*pyDatabaseUrl)
	if err != nil {
		logger.Panic("Can't parse Python database URL.", zap.String("url", *pyDatabaseUrl), zap.Error(err))
	}

	pyDatabase, err := openSqliteDb(*oldSqliteUrl)
	if err != nil {
		logger.Panic("Couldn't open Python database.", zap.String("url", *pyDatabaseUrl), zap.Error(err))
	}
	defer pyDatabase.Close()

	err = process(*pyDatabase, goDatabase)
	if err != nil {
		logger.Error("Error while processing data", zap.Error(err))
	}
}

func process(pyDb sql.DB, goDb persistence.Database) error {
	torrentsTotal, err := getNumberOfTorrents(pyDb)
	if err != nil {
		zap.L().Panic("Couldn't count torrents", zap.Error(err))
	}

	bar := pb.New(int(torrentsTotal))
	//bar.SetRefreshRate(time.Second)
	bar.Start()

	// Selecting torrents
	tRows, err := pyDb.Query("SELECT id, info_hash, name, total_size, discovered_on FROM torrents ORDER BY id ASC;")
	if err != nil {
		zap.L().Fatal("Error when querying torrents from old database.")
	}
	defer tRows.Close()

	torrentLoop:
	for tRows.Next() {
		var torrent persistence.TorrentMetadata

		bar.Increment()

		err := tRows.Scan(&torrent.ID, &torrent.InfoHash, &torrent.Name, &torrent.Size, &torrent.DiscoveredOn)
		if err != nil {
			zap.L().Error("Error scanning torrent row.", zap.Error(err))
			continue
		}

		// Selecting files for torrent
		fRows, err := pyDb.Query("SELECT f.path, f.size FROM files AS f WHERE f.torrent_id = ? ORDER BY f.id ASC;", torrent.ID)
		if err != nil {
			zap.L().Fatal("Error when querying files for torrent.", zap.Uint64("torrent_id", torrent.ID))
		}
		defer fRows.Close()

		files := make([]persistence.File, 0)

		for fRows.Next() {
			var file persistence.File
			err := fRows.Scan(&file.Path, &file.Size)
			fRows.Close()
			if err != nil {
				zap.L().Error("Error scanning file row", zap.Uint64("torrent_id", torrent.ID), zap.Error(err))
				continue torrentLoop
			}

			files = append(files, file)
		}

		exists, err := goDb.DoesTorrentExist(torrent.InfoHash)
		if err != nil {
			zap.L().Error("Error checking torrent existence.", zap.String("hash", hex.EncodeToString(torrent.InfoHash)), zap.Error(err))

			continue
		}

		if !exists {
			err = goDb.AddNewTorrent(torrent.InfoHash, torrent.Name, files)
			if err != nil {
				zap.L().Error("Error adding torrent to Go database.", zap.String("hash", hex.EncodeToString(torrent.InfoHash)), zap.Error(err))
			}
		}
	}

	bar.Finish()

	return nil
}

func getNumberOfTorrents(db sql.DB) (uint, error) {
	rows, err := db.Query("SELECT COUNT(id) FROM torrents;")
	if err != nil {
		return 0, err
	}
	defer rows.Close()

	if rows.Next() != true {
		return 0, fmt.Errorf("No rows returned from `SELECT COUNT(id)`")
	}

	var n *uint
	if err = rows.Scan(&n); err != nil {
		return 0, err
	}

	// If the database is empty (i.e. 0 entries in 'torrents') then the query will return nil.
	if n == nil {
		return 0, nil
	} else {
		return *n, nil
	}
}

func openSqliteDb(url_ url.URL) (*sql.DB, error) {
	url_.Scheme = ""
	return sql.Open("sqlite3", url_.String())
}
